import json
import urllib3
from concurrent.futures import as_completed
from requests_futures.sessions import FuturesSession
from loguru import logger
import requests
import pkgutil
import time
from .helpers import _get_morpheus_license_from_cypher

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def setup_aio(node_ip,account_name,user_name,password,email,license):
    """
    The setup_aio function is used to configure a Morpheus Appliance.
    It will take the following parameters:
        node_ip - The IP address of the appliance you wish to configure.
        account_name - The name of the account you wish to create on this appliance. This should be all lowercase letters and numbers only, no special characters or spaces. 
        user_name - The username for this new account (should match your email address). This should be all lowercase letters and numbers only, no special characters or spaces.
     
    :param node_ip: Determine the ip address of the node to be configured
    :param account_name: Set the account name for the new user
    :param user_name: Set the admin user name
    :param password: Set the password for the first user created
    :param email: Set the email address for the admin user
    :param license: Pass in the license file
    :return: A string
    :doc-author: Trelent
    """
    logger.info(f'Begin attempt to set up Morpheus AIO')
    try:
        ping = ""
        timeout = 1
        while "MORPHEUS PING" not in ping:
            logger.info(f'Timeout count now at {timeout}')
            try:
                ping = _get_appliance_node_status(node_ip)
                if "Morpheus is Loading..." in ping:
                    logger.info(f'{ping} checking again in 60 seconds')
                    time.sleep(60)
                    timeout = timeout + 1
                elif "404" in str(ping):
                    logger.info(f'Encountered a 404. Assuming it is still initializing. Waiting 60 seconds to check again.')
                    time.sleep(60)
                    timeout = timeout + 1
                elif "Could not reach node:" in ping:
                    logger.info(f'Unable to reach the node. Assuming it is still initializing. Waiting 60 seconds to check again.')
                    time.sleep(60)
                    timeout = timeout + 1
                elif timeout >= 45:
                    return("Timout waiting for the system to come up")
            except Exception as e:
                logger.error(f'Unhandled exception occurred')
                return(e)
        try:
            logger.info(f'Attempting to configure with the follow configuration:')
            _apply_morpheus_setup_config(node_ip,account_name,user_name,password,email)
        except Exception as e:
            logger.error(f'Unhandled exception: {e}')
            return(e)
        try:
            token = get_morpheus_api_token(node_ip,user_name,password)
        except Exception as e:
            logger.error(f'Unhandled exception: {e}')
            return(e)
        try:
            _apply_morpheus_license(node_ip,token,license)
        except Exception as e:
            logger.error(f'Unhandled exception: {e}')
            return(e)
    except Exception as e:
        logger.error(f'Unhandled exception occurred: {e}')


def setup_three_node():
    pass

def _apply_morpheus_setup_config(node_ip,account_name,user_name,password,email):
    """
    The _apply_morpheus_setup_config function is used to configure the Morpheus Appliance. 
    It takes in a node_ip, account_name, user_name, password and email as arguments. It then performs a POST request to the /api/setup/init endpoint on that node using those credentials.
    
    :param node_ip: Specify the ip address of the node to configure
    :param account_name: Set the morpheus appliance name
    :param user_name: Specify the username that will be used to login to morpheus
    :param password: Set the password for the initial user account
    :param email: Provide a default email address to be used in the initial configuration of morpheus
    :return: The response from the api call
    :doc-author: Trelent
    """
    logger.info("Begin attempt to configure Morpheus")
    url = f'https://{node_ip}'
    endpoint = "/api/setup/init"
    headers={'Content-Type': 'application/json',"Accept":"application/json"}
    body={"applianceUrl": url, "applianceName": account_name, "accountName": account_name, "username": user_name, "password": password, "email": email, "firstName": user_name }
    try:
        logger.info(f'Attempting to perform first time setup on node {node_ip}')
        resp = requests.post(f'{url}{endpoint}',headers=headers,verify=False,data=json.dumps(body))
    except Exception as e:
        logger.error(f'Something went wrong: {e}')
    if "200" in str(resp):
        logger.info(f'Initial configuration of Morpheus successful')
    else:
        logger.error(f'It appears something went wrong')
    return(resp)

def _get_appliance_node_status(node_ip):
    """
    The _get_appliance_node_status function is used to check if the Morpheus UI is up on a given node.
    It takes in a single argument, which is the IP address of the node you want to check. It returns either
    the string &quot;Morpheus UI Up&quot; or an error message.
    
    :param node_ip: Pass the ip address of the node to be checked
    :return: The response from the node
    :doc-author: Trelent
    """
    headers = {'Content-Type': 'application/json'}
    url = f'https://{node_ip}'
    endpoint = "/ping"
    try:
        logger.info(f'Checking for Morpheus UI to be up: {node_ip}')
        resp = requests.get(f'{url}{endpoint}',headers=headers,verify=False)
    except Exception as e:
        logger.error(f'Could not reach node: {node_ip}')
        return(f'Could not reach node: {node_ip}')
    if "MORPHEUS PING" in str(resp.text):
        out = resp.text
        logger.info(f'Looks like the Morpehus UI is up: {out}')
    elif "Morpheus is Loading..." in str(resp.text):
        out = "Morpheus is Loading..."
    else:
        out = resp
    return(out)

def get_morpheus_api_token(url,user_name,password):
    """
    The get_morpheus_api_token function is used to get a Morpheus API token.
    It takes three arguments: the URL of the Morpheus appliance, a username and password.
    The function returns an access_token which can be used in subsequent requests.
    
    :param url: Specify the morpheus api instance that is being used
    :param user_name: Specify the username of the user that is used to authenticate with the morpheus api
    :param password: Pass the password to get_morpheus_api_token function
    :return: A token for the user
    :doc-author: Trelent
    """
    if 'https://' not in url:
        url = f'https://{url}'
    endpoint = "/oauth/token?grant_type=password&scope=write&client_id=morph-api"
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}
    body = {"username": user_name, "password": password}
    try:
        logger.info(f'Attempting to get API token')
        resp = requests.post(f'{url}{endpoint}',headers=headers,verify=False,data=body)
    except Exception as e:
        logger.info(f'Something went wrong')
    if "200" in str(resp):
        logger.info("Token acquired")
        token = resp.json()["access_token"]
        return(token)
    elif "400" in str(resp):
        logger.error('Bad credentials')
        return(resp)
    else:
        return(resp)
    
def _apply_morpheus_license(url,token,license):
    """
    The _apply_morpheus_license function is used to apply a license to the Morpheus instance.
    It takes three arguments: url, token, and license. The url argument is the URL of your Morpheus instance (e.g., https://morpheus-instance). 
    The token argument is an API key generated in the UI for that particular instance (e.g., 1234567890abcde1234567890abcde1234567890). 
    The license argument should be a string containing your desired morpheus license.
    
    :param url: Specify the morpheus instance
    :param token: Pass the token generated by the _get_token function
    :param license: Pass in the license key that you want to apply
    :return: A response object
    :doc-author: Trelent
    """
    headers={'Content-Type': 'application/json',"Accept":"application/json", "Authorization": "BEARER " + token}
    if 'https://' not in url:
        url = f'https://{url}'
    endpoint = "/api/license"
    body = {"license": license}
    try:
        logger.info(f'Attempting to apply the license')
        resp = requests.post(f'{url}{endpoint}',headers=headers,verify=False,data=json.dumps(body))
    except Exception as e:
        logger.error(f'Something went wrong')
    if "200" in str(resp):
        logger.info("License successfully applied")
        return(resp)
    elif "\"success\":false," in str(resp.text):
        logger.error(resp.json()["msg"])
    elif "\"error\":" in str(resp.text):
        logger.error(resp.json()["error_description"])
    return(resp)

def _get_morpheus_setup_status(url):
    """
    The _get_morpheus_setup_status function checks to see if the Morpheus platform is setup.
    It returns True if the platform is setup, and False otherwise.
    
    :param url: Specify the url of the node to check
    :return: A boolean value
    :doc-author: Trelent
    """
    """
    The _get_morpheus_setup_status function checks to see if the Morpheus platform is setup.
    It returns True if the platform is setup, and False otherwise.
    
    :param url: Specify the url of the node to check
    :return: A boolean value
    :doc-author: Trelent
    """
    # Returns True if the platform is setup. Otherwise returns False.
    if 'https://' not in url:
        url = f'https://{url}'
    endpoint = "/api/ping"
    headers = {'Content-Type': 'application/json'}
    try:
        logger.info(f'Checking the setup status of node: {url}')
        resp = requests.get(f'{url}{endpoint}',headers=headers, verify=False)
    except Exception as e:
        logger.error(f'Could not reach node: {url}')
    if "\'setupNeeded\': True," in resp:
        return(False)
    else:
        return(True)
    
def get_aio_external_ips(url,headers,ids:list):
    """
    The get_aio_external_ips function will return a list of IP addresses for all AIO nodes in the cluster.
    This function is used to determine which node should be used as the primary node for license assignment.
    
    :param url: Specify the url of the target cluster
    :param headers: Pass the api key to the get_aio_external_ips function
    :param ids:list: Pass in a list of ids to check for
    :return: The ip addresses of the aio nodes
    :doc-author: Trelent
    """
    session = FuturesSession()
    endpoint = "/api/instances"
    ips = []
    futures=[session.get(f'{url}{endpoint}/{i}',headers=headers,verify=False) for i in ids]
    logger.info(f'Checking for AIO nodes')
    for future in as_completed(futures):
        resp = future.result()
        if "404" in str(resp):
            logger.error(f'Returned 404')
            return(resp)  
        elif "400" in str(resp):
            logger.error(f'Returned 400')
            return(resp) 
        elif "200" in str(resp):
            for node in resp.json()["instance"]["containerDetails"]:
                if "single-node" in node["externalHostname"]:
                    logger.info(f'Found AIO node IP of : {node["ip"]}')
                    ips.append(node["ip"])
    return(ips)

def get_lab_info(url,headers,ids:list):
    """
    The get_lab_info function accepts a URL, headers and instance ids as arguments.
    It then makes an API call to the Morpheus API to get information about each instance in the lab.
    The function returns a list of lists containing all of the data for each instance.
    
    :param url: Specify the endpoint of the api
    :param headers: Pass the api key to the get_lab_info function
    :param ids:list: Pass in the list of instance ids to be parsed
    :return: A list of lists that can be used to create a table
    :doc-author: Trelent
    """
    session = FuturesSession()
    endpoint = "/api/instances"
    data = [['Lab Name','Region','AIO External IP','AIO Internal IP','AMI ID','Access Key','Secret Key']]
    futures=[session.get(f'{url}{endpoint}/{i}/state',headers=headers,verify=False) for i in ids]
    lb_endpoint = []
    rds_endpoint = []
    for future in as_completed(futures):
        resp = future.result()
        if "200" in str(resp):
            # logger.info(f'Parsing instance state for instance id: {resp.json()["workloads"][0]["refId"]}')
            # logger.info(f'Parsing the terraform resources')
            instance_data = []
            lab_name = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "lab_name"]
            region = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "region"]
            eip = [i["value"]["value"][0][0]["public_ip"] for i in resp.json()["output"]["outputs"] if i["name"] == "elastic_ips"]
            private_ip = [i["value"]["value"][0][0]["private_ip"] for i in resp.json()["output"]["outputs"] if i["name"] == "elastic_ips"]
            ami_id = [i["value"]["value"][0][0]["ami"] for i in resp.json()["output"]["outputs"] if i["name"] == "instances"]
            state_data = json.loads(resp.json()["stateData"])
            access_key = [i["values"]["id"] for i in state_data["values"]["root_module"]["child_modules"][0]["resources"] if i["type"] == "aws_iam_access_key"]
            secret_key = [i["values"]["secret"] for i in state_data["values"]["root_module"]["child_modules"][0]["resources"] if i["type"] == "aws_iam_access_key"]
            instance_data.extend([lab_name[0],region[0],eip[0],private_ip[0],ami_id[0],access_key[0],secret_key[0]])
            try: 
                lb_endpoint = [i["value"]["value"][0]["dns_name"] for i in resp.json()["output"]["outputs"] if i["name"] == "load_balancer"]
                data[0].append('LB Endpoint')
                instance_data.append(lb_endpoint[0])
            except:
                pass
            try:
                rds_endpoint = [i["value"]["value"][0][0]["endpoint"] for i in resp.json()["output"]["outputs"] if i["name"] == "rds"]
                data[0].append('RDS Endpoint')
                instance_data.append(rds_endpoint[0])
            except:
                pass
            try:
                instances = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "instances"]
                app_nodes = instances[0][1]
                app_1_pub = [i["public_ip"] for i in app_nodes if "APP-01" in i["tags_all"]["Name"]]
                data[0].append('App-01 Public IP')
                instance_data.append(app_1_pub[0])
                app_1_priv = [i["private_ip"] for i in app_nodes if "APP-01" in i["tags_all"]["Name"]]
                data[0].append('App-01 Private IP')
                instance_data.append(app_1_priv[0])
                app_2_pub = [i["public_ip"] for i in app_nodes if "APP-02" in i["tags_all"]["Name"]]
                data[0].append('App-02 Public IP')
                instance_data.append(app_2_pub[0])
                app_2_priv = [i["private_ip"] for i in app_nodes if "APP-02" in i["tags_all"]["Name"]]
                data[0].append('App-02 Private IP')
                instance_data.append(app_2_priv[0])
                app_3_pub = [i["public_ip"] for i in app_nodes if "APP-03" in i["tags_all"]["Name"]]
                data[0].append('App-03 Public IP')
                instance_data.append(app_3_pub[0])
                app_3_priv = [i["private_ip"] for i in app_nodes if "APP-03" in i["tags_all"]["Name"]]
                data[0].append('App-03 Private IP')
                instance_data.append(app_3_priv[0])
            except:
                pass
            try:
                instances = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "instances"]
                perc_nodes = instances[0][2]
                perc_1_pub = [i["public_ip"] for i in perc_nodes if "PERC-01" in i["tags_all"]["Name"]]
                data[0].append('Perc-01 Public IP')
                instance_data.append(perc_1_pub[0])
                perc_1_priv = [i["private_ip"] for i in perc_nodes if "PERC-01" in i["tags_all"]["Name"]]
                data[0].append('Perc-01 Private IP')
                instance_data.append(perc_1_priv[0])
                perc_2_pub = [i["public_ip"] for i in perc_nodes if "PERC-02" in i["tags_all"]["Name"]]
                data[0].append('Perc-02 Public IP')
                instance_data.append(perc_2_pub[0])
                perc_2_priv = [i["private_ip"] for i in perc_nodes if "PERC-02" in i["tags_all"]["Name"]]
                data[0].append('Perc-02 Private IP')
                instance_data.append(perc_2_priv[0])
                perc_3_pub = [i["public_ip"] for i in perc_nodes if "PERC-03" in i["tags_all"]["Name"]]
                data[0].append('Perc-03 Public IP')
                instance_data.append(perc_3_pub[0])
                perc_3_priv = [i["private_ip"] for i in perc_nodes if "PERC-03" in i["tags_all"]["Name"]]
                data[0].append('Perc-03 Private IP')
                instance_data.append(perc_3_priv[0])
            except:
                pass
            data.append(instance_data)
    return(data)

def get_lab_instance_ids_from_tag(url,headers,controller_id,tag_name):
    """
    The get_lab_instance_ids_from_tag function accepts a Morpheus API URL,
    a dictionary of HTTP headers containing the user's bearer token,
    the ID of the controller instance to query for lab instances, and 
    the name of the tag that contains all lab instance IDs. The function returns a list 
    of all lab instance IDs.
    
    :param url: Specify the controller url
    :param headers: Pass the api key to the get_lab_instance_ids function
    :param controller_id: Identify the controller to which the api call will be made
    :param tag_name: Specify the tag that you want to search for
    :return: A list of instance ids that are tagged with the specified tag name
    :doc-author: Trelent
    """
    endpoint = "/api/instances"
    resp = requests.get(f'{url}{endpoint}/{controller_id}',headers=headers,verify=False)
    tags = resp.json()["instance"]["tags"]
    for tag in tags:
        if tag["name"] == tag_name:
            ids = tag["value"]
    return(json.loads(ids))






import json
import urllib3
from concurrent.futures import as_completed
from requests_futures.sessions import FuturesSession
import requests
import pkgutil
from loguru import logger
import time

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def get_class_lab_instance_type(url,headers,class_name):
    try:
        logger.info(f'Attempting to get the instance type id by code: {class_name}')
        resp = requests.get(f'{url}/api/library/instance-types?code={class_name}',headers=headers,verify=False)
    except Exception as e:
        logger.error(f'Something went wrong: {e}')
    if "200" in str(resp):
        logger.info(f'Instance type found')
        logger.info(f'ID: {resp.json()["instanceTypes"][0]["id"]}')
        instance_type = resp.json()["instanceTypes"][0]
        return(instance_type)
    else:
        logger.error(f'Instance type not found')

def get_class_layout_id(url,headers,custom_options,instance_type_map):
    endpoint = "/api/library/layouts"
    for l in instance_type_map["instanceTypeLayouts"]:
        try:
            logger.info(f'Attempting to get the layout with the ID of: {l["id"]}')
            resp = requests.get(f'{url}{endpoint}/{l["id"]}', headers=headers, verify=False)
        except Exception as e:
            logger.error(f'Something went horribly wrong here: {e}')
        if "200" in str(resp):
            logger.info(f'Layout found')
            logger.info(f'ID: {resp.json()["instanceTypeLayout"]["id"]}')
            logger.info(f'Version: {resp.json()["instanceTypeLayout"]["instanceVersion"]}')
            logger.info('Comparing to class version...')
            if resp.json()["instanceTypeLayout"]["instanceVersion"] == custom_options["class_version"]:
                logger.info(f'Version match found')
                layout_id = resp.json()["instanceTypeLayout"]["id"]
                return(layout_id)
            else:
                logger.info(f'Version mismatch. Skipping...')

def get_morpheus_terraform_plan_id(url,headers):
    """
    The get_morpheus_terraform_plan_id function is used to get the id of the Morpheus Terraform plan.
    This function returns a single integer value which represents the ID of the plan.
    
    :param url: Specify the morpheus api url
    :param headers: Pass the token generated by morpheus
    :return: The id of the morpheus terraform plan
    :doc-author: Trelent
    """
    endpoint = "/api/service-plans?phrase=Terraform"
    try:
        logger.info(f'Attempting to get the Morpheus Terraform plan id.')
        resp = requests.get(f'{url}{endpoint}', headers=headers, verify=False).json()
    except Exception as e:
        logger.error(f'Something went horribly wrong {e}')
    return(resp["servicePlans"][0]["id"])

def get_instance_provisioning_payload(class_version,zone_id,instance_name,site_id,instance_type,instance_type_code,layout_id,plan_id,template_parameters):
    """
    The get_instance_provisioning_payload function takes the provided variables and loads a template file that contains
    the payload for the instance provisioning API call. The function then updates the template with all of the provided 
    variables, and returns it as a JSON object.
    
    :param zone_id: Identify the zone in which to create the instance
    :param instance_name: Name the instance
    :param site_id: Determine the site to which the instance will be attached
    :param instance_type: Determine the instance type code for the payload
    :param instance_type_code: Determine the type of instance to be provisioned
    :param layout_id: Specify the layout of the instance
    :param plan_id: Determine the instance type
    :param template_parameters: Pass in the parameters that are needed to provision an instance
    :return: A payload template that can be used to provision an instance
    :doc-author: Trelent
    """
    try:
        logger.info(f'Attempting to load the payload template file')
        f = pkgutil.get_data(__name__, 'template_files/provisioning_payload.json')
        f = json.loads(f)
    except Exception as e:
        logger.error(f'Something has gone awry. {e}')
    try:
        logger.info(f'Updating the payload template with the provided variables')
        f["zoneId"] = zone_id
        f["instance"]["name"] = instance_name
        f["instance"]["site"]["id"] = site_id
        f["instance"]["type"] = instance_type
        f["instance"]["instanceType"]["code"] = instance_type_code
        f["instance"]["layout"]["id"] = layout_id
        f["instance"]["plan"]["id"] = plan_id
        f["config"]["templateParameter"] = template_parameters
    except Exception as e:
        logger.error(f'Terrible failure herein {e}')
    return(f)

def get_instance(url,headers,instance_id):
    """
    The get_instance function is used to get the details of a specific instance. It does this by making an API call
    to the Morpheus API, and then returning the response as a JSON object.
    
    :param url: Specify the morpheus api url
    :param headers: Pass the token generated by morpheus
    :param instance_id: Pass the id of the instance to get
    :return: A JSON object containing the details of the instance
    :doc-author: Trelent
    """
    endpoint = "/api/instances"
    try:
        logger.info(f'Attempting to get the instance with the ID of: {instance_id}')
        resp = requests.get(f'{url}{endpoint}/{instance_id}', headers=headers, verify=False)
    except Exception as e:
        logger.error(f'Something went horribly wrong {e}')
        return(e)
    if "200" in str(resp):
        logger.info(f'Instance found')
        return(resp.json()["instance"])
    else:
        logger.error(f'Instance not found')

def gen_instance_names(instance_map):
    start = 0
    end = instance_map["customOptions"]["seatCount"]
    logger.info(f'Generating a list of instance names')
    list = []
    while start < end:
        start += 1
        list.append(f'{instance_map["customOptions"]["class_name"][0:3]}-{instance_map["customOptions"]["trainingZone"].split(",")[1][0:2]}{instance_map["customOptions"]["trainingZone"].split(",")[1].split("-")[1][0:1]}{instance_map["customOptions"]["trainingZone"].split(",")[1][-1]}-{instance_map["customOptions"]["lab_name"]}{start:02d}'.format(start))
    return(list)

def get_morpheus_terraform_inputs(morpheus_custom_options):
    """
    The get_lab_terraform_inputs function takes the provided Morpheus custom options and parses them to determine which template parameters need to be filled in. It then fills in the appropriate values for those parameters and returns a dictionary of all of the template parameters with their corresponding values.
    
    :param morpheus_custom_options: Pass in the custom options that have been provided by the user
    :return: A dictionary of the key value pairs that should be used for the lab
    :doc-author: Trelent
    """
    # Set the class version
    class_version = morpheus_custom_options["class_version"]

    # Validate the class type and import the correct file template for the template parameters to fill in
    if "administration" in morpheus_custom_options["class_name"]:
        logger.info(f'Administration class selected. Importing admin_class_tf_vars.json')
        f = pkgutil.get_data(__name__, 'template_files/admin_class_tf_vars.json')
        f = json.loads(f)
    elif "installation" in morpheus_custom_options["class_name"]:
        logger.info(f'Installation class selected. Importing instal_class_tf_vars.json')
        f = pkgutil.get_data(__name__, 'template_files/install_class_tf_vars.json')
        f = json.loads(f)
    elif 'automation' in morpheus_custom_options["class_name"]:
        logger.info(f'Automation class selected. Importing automation_class_tf_vars.json')
        f = pkgutil.get_data(__name__, 'template_files/automation_class_tf_vars.json')
        f = json.loads(f)
    elif 'troubleshooting' in morpheus_custom_options["class_name"]:
        logger.info(f'Troubleshooting class selected. Importing troubleshooting_class_tf_vars.json')
        f = pkgutil.get_data(__name__, 'template_files/troubleshooting_class_tf_vars.json')
        f = json.loads(f)
    terraform_inputs = {}
    # Parse the key value pairs of the provided Morpheus custom options
    for k, v in morpheus_custom_options.items():
        logger.info(f'Checking for a match on the option type: {k}')
        try: 
            if k in f:
                logger.info(f'Found {k}, updating with a value of {v}')
                update = {k: v}
                terraform_inputs.update(update)
            else:
                logger.info(f'{k} not found in the template. Skipping it.')
        except Exception as e:
            logger.error(f'Something went wrong {e}')
    for k, v in f.items():
        try:
            if v :
                logger.info(f'Item {k} with a value of {v} discovered. Adding it to the payload.')
                update = {k: v}
                terraform_inputs.update(update)
            else:
                logger.info(f'{k} does not have a default value in the template. Skipping it')
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
    return(terraform_inputs)

def extract_json_content(text):
    start_token = "----json_object_start----"
    end_token = "----json_object_end----"
    
    start_index = text.find(start_token)
    end_index = text.find(end_token)
    
    if start_index == -1 or end_index == -1 or start_index >= end_index:
        return None  # Return None if the start or end token is not found, or if start comes after end

    return text[start_index + len(start_token):end_index].strip()


import json
import urllib3
from concurrent.futures import as_completed
from requests_futures.sessions import FuturesSession
import requests
import pkgutil
from loguru import logger
import time
from morpheuscypher import Cypher
import sys
import os

def _get_instance_ids_from_names(url,headers,names:list):
    """
    The _get_instance_ids_from_names function accepts a list of names and returns a list of instance IDs.
    It does this by making an API call to the AS API for each name in the provided list, and then returning
    the ID from that response.
    
    :param url: Specify the url of the instance
    :param headers: Pass the api key to the _get_instance_ids_from_names function
    :param names:list: Pass a list of names to the _get_instance_ids_from_names function
    :return: A list of instance ids based on a list of names
    :doc-author: Trelent
    """
    session = FuturesSession()
    endpoint = "/api/instances"
    ids = []
    futures=[session.get(f'{url}{endpoint}?name={n}',headers=headers,verify=False) for n in names]
    for future in as_completed(futures):
        resp = future.result()
        if "200" in str(resp):
            i = resp.json()["instances"][0]
            ids.append(i["id"])
    return(ids)

def _get_morpheus_license_from_cypher(url,token,cypher_name):
    """
    The _get_morpheus_license_from_cypher function is a helper function that retrieves the Morpheus License from the Cypher
    database.  It accepts three parameters: url, token, and cypher_name.  The url parameter is required for making an API call to
    the Morpheus service.  The token parameter is required for making an API call to the Morpheus service.  The cypher_name
    parameter specifies which Cypher query will be used in order to retrieve license information from the database.
    
    :param url: Specify the morpheus api url
    :param token: Authenticate the user to morpheus
    :param cypher_name: Specify which cypher query to run
    :return: The license information from the morpheus api
    :doc-author: Trelent
    """
    logger.info(f'Begin get_morpheus_license_from_cypher')
    c = Cypher(url=url,token=token,ssl_verify=False)
    out = c.get(cypher_name)
    return(out)

def _get_code_repo_id_by_integration_name(url,headers,integration_name):
    # TODO: Run trelent
    _validate_url(url)
    logger.info(f'Searching for an integration by the name of: {integration_name}')
    e1 = "/api/integrations"
    e2 = "/api/options/codeRepositories"
    try:
        resp = requests.get(f'{url}{e1}?name={integration_name}',headers=headers, verify = False)
        if "200" in str(resp):
            i = resp.json()["integrations"]
            logger.info("Checking the count of integrations with that name")
            if len(i) > 1:
                logger.error('There is more than one integration with that name. Not sure what to do. Exiting...')
                return
            if len(i) < 1:
                logger.info(f'There is no integration by the name: {integration_name}')
                return(0)
            if i[0]["type"] != "git":
                logger.error(f'This is not a git based integration. No code repo to find')
                return
            logger.info(f'Collecting the name of the code repository')
            cr = i[0]["url"].split("/")[-1]
            integration_id = i[0]["id"]
            try:
                logger.info(f'Attempting to make the API call for: {url}{e2}?integrationId={integration_id}')
                resp = requests.get(f'{url}{e2}?integrationId={integration_id}',headers=headers, verify=False)
                if "200" in str(resp):
                    logger.info("Success!")
                    i = resp.json()["data"]
                    if len(i) > 1:
                        logger.info('Found more than one code repo this url...thinking....')
                        for r in i:
                            if integration_name in r:
                                data = r["value"]
                    code_repo_id = i[0]["value"] 
                else:
                    data = resp
            except Exception as e:
                logger.error(f'Encountered an error: {e}')
                return
    except Exception as e:
        logger.error(f'Encountered an error: {e}')
        return
    integration_info = {
            "integration_id": integration_id,
            "linked_code_repo_id": code_repo_id
            }
    return(integration_info)

def _add_git_integration_pat_auth(url,headers,integration_name,git_url,git_branch,git_username,git_pat):
    logger.info("Begin attempt to add git integration")
    _validate_url(url)
    endpoint = "/api/integrations"
    check = _get_code_repo_id_by_integration_name(url,headers,integration_name)
    if check:
        logger.info(f'A repo with the name {integration_name} already exists')
        return(check)
    logger.info(f'Personal access token selected as the auth method for the git repo')
    f = pkgutil.get_data(__name__, "template_files/git_auth_type_pat.json")
    jbody = json.loads(f)
    jbody["integration"]["name"] = integration_name
    jbody["integration"]["serviceUrl"] = git_url
    jbody["integration"]["config"]["defaultBranch"] = git_branch
    jbody["integration"]["serviceUsername"] = git_username
    jbody["integration"]["servicePassword"] = git_pat
    jbody["integration"]["serviceToken"] = git_pat
    body = json.dumps(jbody)
    try:
        logger.info(f'Attempting to create the Git integration')
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
        data = resp.json()
    except Exception as e:
        logger.error(e)
        return
    if data["success"]:
        logger.info(f'Successfully add GIT integration at id: {data["integration"]["id"]}')
        logger.info(f'Pausing to allow code sync')
        time.sleep(30)
        logger.info(f'Acquiring linked code repo id...')
        code_repo = _get_code_repo_id_by_integration_name(url,headers,integration_name)
        logger.info(f'Acquired the code repo: {code_repo}')
        integration_info = {
            "integration_id": data["integration"]["id"],
            "linked_code_repo_id": code_repo["linked_code_repo_id"]
            }
        return(integration_info)
    else:
        logger.error(f'Something went wrong')
        logger.error(resp.text)
        return(resp.text)

def _get_option_list_id_by_name(url,headers,name):
    logger.info(f'Begin attempt to check for existing option list: {name}')
    endpoint = "/api/library/option-type-lists"
    resource_type = "optionTypeLists"
    try:
        resp = requests.get(f'{url}{endpoint}?name={name}', headers=headers, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        logger.info(f'API call successful. Parsing results....')
        resp = resp.json()
        if len(resp[resource_type]) == 1:
            logger.info(f'Found one {resource_type} with name {name} at id: {resp[resource_type][0]["id"]}')
            logger.info(f'Returning ID')
            return(resp[resource_type][0]["id"])
        if len(resp[resource_type]) == 0:
            logger.info(f'No {resource_type} by the name found')
            return 0
        if len(resp[resource_type]) > 1:
            logger.info(f'More than one {resource_type} by that name exists...not sure what to do.')
            exit(0)

def _add_manual_option_list(url,headers,name):
    # TODO: Set this up for multiple lists or manual input of initial dataset
    logger.info(f'Attempting to create the manual option list {name}')
    endpoint = "/api/library/option-type-lists"
    resource_type = "optionTypeList"
    check = _get_option_list_id_by_name(url,headers,name)
    if check:
        logger.info(f'A {resource_type} with the name {name} already exists')
        logger.info(f'Returning the ID')
        return(check)
    file_template = f'template_files/manual_option_list_{name.replace(" ","_").lower()}.json'
    logger.info(f'Attempting to load file template: {file_template}')
    f = pkgutil.get_data(__name__, file_template)
    ds = f.decode('utf-8')
    jbody = {
        "optionTypeList": {
            "name": name,
            "type": "manual",
            "visibility": "private",
            "initialDataset": ds
      }
    }
    body = json.dumps(jbody)
    try:
        resp = requests.post(f'{url}{endpoint}',headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created the {resource_type} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return(data[resource_type]["id"])

def _get_input_id_by_name(url,headers,input_name):
    logger.info(f'Checking for input with name: {input_name}')
    endpoint = "/api/library/option-types"
    resource_type = "optionTypes"
    try:
        resp = requests.get(f'{url}{endpoint}?name={input_name}', headers=headers, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        logger.info(f'API call successful. Parsing results....')
        resp = resp.json()
        if len(resp[resource_type]) == 1:
            logger.info(f'Found one {resource_type} input with name {input_name} at id: {resp[resource_type][0]["id"]}')
            logger.info(f'Returning ID')
            return(resp[resource_type][0]["id"])
        if len(resp[resource_type]) == 0:
            logger.info(f'No {resource_type} by the name found')
            return 0
        if len(resp[resource_type]) > 1:
            logger.info(f'More than one {resource_type} by that name exists...not sure what to do.')
            exit(0)

def _add_input(url,headers,input_name, field_name, input_type, field_label, required):
    logger.info(f'Begin attempt to add input: {input_name}')
    endpoint = "/api/library/option-types"
    resource_type = "optionType"
    check = _get_input_id_by_name(url,headers,input_name)
    if check:
      logger.info(f'Found one {resource_type} by that name. Using it\'s id')
      return check
    jbody = {
      "optionType": {
        "name": input_name,
        "fieldName": field_name,
        "type": input_type,
        "fieldLabel": field_label,
        "required": required,
        "exportMeta": False
      }
    }
    body = json.dumps(jbody)
    try:
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created {resource_type} {input_name} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return data[resource_type]["id"]

def _add_select_input(url, headers, input_name, field_name, option_list_id, field_label, required):
    logger.info(f'Begin attempt to add input: {input_name}')
    endpoint = "/api/library/option-types"
    resource_type = "optionType"
    check = _get_input_id_by_name(url,headers,input_name)
    if check:
      logger.info(f'Found one {resource_type} by that name. Using it\'s id')
      return check
    jbody = {
      "optionType": {
        "name": input_name,
        "fieldName": field_name,
        "type": "select",
        "optionList": {
          "id": option_list_id
        },
        "fieldLabel": field_label,
        "required": required,
        "exportMeta": False
      }
    }
    body = json.dumps(jbody)
    try:
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created {resource_type} {input_name} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return data[resource_type]["id"]

def _get_task_id_by_name(url,headers,name):
    logger.info(f'Begin attempt to get task id by name')
    endpoint = "/api/tasks"
    resource_type = "tasks"
    try:
        resp = requests.get(f'{url}{endpoint}?name={name}',headers=headers,verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        logger.info(f'API call successful. Parsing results...')
        resp = resp.json()
        if len(resp[resource_type]) == 1:
            logger.info(f'Found one {resource_type} with name {name} at id: {resp[resource_type][0]["id"]}')
            logger.info(f'Returning ID')
            return(resp[resource_type][0]["id"])
        if len(resp[resource_type]) == 0:
            logger.info(f'No {resource_type} by the name {name} found')
            return 0
        if len(resp[resource_type]) > 1:
            logger.info(f'More than one {resource_type} by that name exists...not sure what to do.')
            exit(0)


def _add_python_task(url,headers,name, code, task_type_id, script_code, result_type, source_type, repo_id, content_path, pythonargs, packages):
    logger.info(f'Begin attempt to create python task')
    endpoint = "/api/tasks"
    resource_type = "task"
    check = _get_task_id_by_name(url,headers,name)
    if check:
      logger.info(f'Found one {resource_type} by that name. Using it\'s id')
      return check
    jbody = {
      "task": {
        "name": name,
        "code": code,
        "taskType": {
          "id": task_type_id,
          "code": script_code
        },
        "resultType": result_type,
        "file": {
          "sourceType": source_type,
          "repository": {
            "id": repo_id
          },
          "contentPath": content_path,
          "contentRef": "master"
        },
        "taskOptions": {
          "pythonArgs": pythonargs,
          "pythonAdditionalPackages": packages
        },
        "executeTarget": "local",
        "retryable": False,
        "allowCustomConfig": False
      }
    }
    body = json.dumps(jbody)
    try:
        logger.info(f'URL: {url}{endpoint}')
        logger.info(f'Headers: {headers}')
        logger.info(f'Body: {body}')
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created {resource_type} {name} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return data[resource_type]["id"]
    else:
        logger.error(f'API call failed with response: {resp}')
        logger.error(resp.text)
        exit(0)

def _get_workflow_id_by_name(url,headers,name):
    logger.info(f'Begin attempt to get workflow by name')
    endpoint = "/api/task-sets"
    resource_type = "taskSets"
    try:
        resp = requests.get(f'{url}{endpoint}?name={name}',headers=headers,verify=False)
    except Exception as e:
        logger.error(e)
    if "200" in str(resp):
        logger.info(f'API call successful. Parsing results...')
        resp = resp.json()
        if len(resp["taskSets"]) == 1:
            logger.info(f'Found one {resource_type} with name {name} at id: {resp[resource_type][0]["id"]}')
            logger.info(f'Returning ID')
            return(resp[resource_type][0]["id"])
        if len(resp[resource_type]) == 0:
            logger.info(f'No {resource_type} by the name found')
            return 0
        if len(resp[resource_type]) > 1:
            logger.info(f'More than one {resource_type} by that name exists...not sure what to do.')
            exit(0)

def _add_operational_workflow(url,headers,name, taskid, inputs):
    # TODO: Get multiple tasks working
    logger.info(f'Begin attempt to create workflow')
    endpoint = "/api/task-sets"
    resource_type = "taskSet"
    check = _get_workflow_id_by_name(url,headers,name)
    if check:
      logger.info(f'Found one {resource_type} by that name. Using it\'s id')
      return check
    jbody = {
      "taskSet": {
        "name": name,
        "type": "operation",
        "tasks": [
          {
            "taskId": taskid,
            "taskPhase": "operation"
          }
        ],
        "optionTypes": inputs
      }
    }
    body = json.dumps(jbody)
    try:
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created {resource_type} {name} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return data[resource_type]["id"]

def _get_catalog_id_by_name(url,headers,name):
    logger.info(f'Begin attempt to get catalog item by name')
    endpoint = "/api/catalog-item-types"
    resource_type = "catalogItemTypes"
    try:
        resp = requests.get(f'{url}{endpoint}?name={name}',headers=headers,verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        logger.info(f'API call successful. Parsing results...')
        resp = resp.json()
        if len(resp[resource_type]) == 1:
            logger.info(f'Found one {resource_type} with name {name} at id: {resp[resource_type][0]["id"]}')
            logger.info(f'Returning ID')
            return(resp[resource_type][0]["id"])
        if len(resp[resource_type]) == 0:
            logger.info(f'No {resource_type} by the name found')
            return 0
        if len(resp[resource_type]) > 1:
            logger.info(f'More than one {resource_type} by that name exists...not sure what to do.')
            exit(0)

def _add_workflow_catalog_item(url,headers,name, workflow_id):
    logger.info(f'Begin attempt to create catalog item')
    endpoint = "/api/catalog-item-types"
    resource_type = "catalogItemType"
    check = _get_catalog_id_by_name(url,headers,name)
    if check:
      logger.info(f'Found one {resource_type} by that name. Using it\'s id')
      return check
    jbody = {
      "catalogItemType": {
           "type": "workflow",
           "enabled": True,
           "featured": False,
           "workflow": {
                "id": workflow_id
           },
           "name": name,
           "context": "appliance"
      }
    }
    body = json.dumps(jbody)
    try:
        resp = requests.post(f'{url}{endpoint}', headers=headers, data=body, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        data = resp.json()
        logger.info(f'Successfully created {resource_type} {name} at ID: {data[resource_type]["id"]}')
        logger.info(f'Returning {resource_type} ID')
        return data[resource_type]["id"]

def _add_catalog_logo(url,headers,catalog_id, image):
    logger.info(f'Begin attempt to add logo to catalog item')
    endpoint = "/api/catalog-item-types"
    module = pkgutil.get_loader('molab')
    module_path = module.path.rsplit('/',1)[0]
    path  = f'{module_path}/template_files/{image}'
    logger.info(f'Attempting to open file at path: {path}')
    files = {'catalogItemType.logo': open(path, "rb")}
    newheaders = {
        "Authorization": headers["Authorization"],
        "Content-Type": "application/json"
    }
    try:
        resp = requests.put(f'{url}{endpoint}/{catalog_id}/update-logo',headers=newheaders, files=files, verify=False)
    except Exception as e:
        logger.error(e)
        return
    if "200" in str(resp):
        logger.info(f'Successfully uploaded logo')
        return(resp.json())
    else:
        logger.info(f'I don\'t think it worked captain')
        return(resp.text)

def _validate_url(url):
    logger.info('Ensuring https in url')
    if "https://" not in url:
        url = f'https://{url}'
    return(url)

def sleep(time):
    """
    The sleep function is used to pause the program for a specified amount of time.
    It can be useful when waiting for an external service to respond, or when polling
    a database.
    
    :param time: Specify how long the function should sleep for
    :return: None
    :doc-author: Trelent
    """
    logger.info(f'Sleeping for {time}')
    time.sleep(time)


### molab is a python package for setting up and configuring Morpheus training labs ###

# Original Imports
from .deploy import *
from .destroy import *
from .configure import *
from .courses import *

# v6.0 Imports
from .labs import newClass, existingClass, newLab, existingLab, awsEnv
from .tools import gen_instance_names, extract_json_content


import json
import urllib3
from concurrent.futures import as_completed
from requests_futures.sessions import FuturesSession
import requests
import pkgutil
from loguru import logger
from .helpers import _add_git_integration_pat_auth, _validate_url, _add_manual_option_list, _add_input, _add_select_input, _add_python_task, _add_operational_workflow, _add_workflow_catalog_item, _add_catalog_logo


def configure_automation_class(node_ip,headers,**class_configs):
    logger.info(f'Begin attempt to setup Automation Class on node: {node_ip}')
    # Ensure URL is https
    url = _validate_url(node_ip)
    # Add the GIT integration
    integration = _add_git_integration_pat_auth(url,headers,class_configs["integration_name"],class_configs["git_url"],class_configs["git_branch"],class_configs["git_username"],class_configs["git_pat"])
    logger.info(f'Integration Data: {integration}')
    # Add the manual option list
    option_list_id = _add_manual_option_list(url,headers,class_configs["option_list_name"])
    logger.info(f'Option List ID: {option_list_id}')

    ## Create inputs
    input_ids = []
    input_ids.append(_add_input(node_ip,headers,"AMI - CentOS", "amicentos", "text", "AMI - CentOS", "True"))
    input_ids.append(_add_input(node_ip,headers,"AMI - Ubuntu", "amiubuntu", "text", "AMI - Ubuntu", "True"))
    input_ids.append(_add_input(node_ip,headers,"AWS Key", "awsKey", "text", "AWS Key", "True"))
    input_ids.append(_add_input(node_ip,headers,"AWS Secret", "awsSecret", "text", "AWS Secret", "True"))
    input_ids.append(_add_input(node_ip,headers,"AWS VPC", "awsVpc", "text", "AWS VPC", "True"))
    input_ids.append(_add_input(node_ip,headers,"Internal IP", "internalHost", "text", "Internal IP", "True"))
    input_ids.append(_add_select_input(node_ip,headers,"AWS Regions", "awsRegion", option_list_id, "AWS Region", "True"))

    ## Create "Set Up My Lab" Python Task
    task_id = _add_python_task(node_ip,headers,"Set Up My Lab", "setupMyLab", "12", "jythonTask", "value", "repository", integration["linked_code_repo_id"], "/setupmylab.py", "", "requests")

    ## Create "Set Up My Lab" Workflow
    workflow_id = _add_operational_workflow(node_ip,headers,"Set Up My Lab", task_id, input_ids)

    ## Create catalog item
    catalog_id = _add_workflow_catalog_item(node_ip,headers,"Set Up My Lab", workflow_id)

    ## Add logo
    logo = _add_catalog_logo(node_ip,headers,catalog_id, "MorpheusImage.png")


import json, urllib3, requests, pkgutil, time, boto3
from concurrent.futures import as_completed, ThreadPoolExecutor
from requests_futures.sessions import FuturesSession
from loguru import logger
from .tools import get_class_lab_instance_type, get_class_layout_id, get_morpheus_terraform_plan_id, get_instance_provisioning_payload, get_instance, get_morpheus_terraform_inputs
from .helpers import _add_git_integration_pat_auth, _validate_url, _add_manual_option_list, _add_input, _add_select_input, _add_python_task, _add_operational_workflow, _add_workflow_catalog_item, _add_catalog_logo

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class newClass():
    def __init__(self,url,headers,instance_map,instance_name):
        self.url = url
        self.headers = headers
        self.instance_map = instance_map
        self.instanceName = instance_name
        self.instance_id = ""
        self.class_name = self.instance_map["customOptions"]["class_name"]
        self.custom_options = self.instance_map["customOptions"]
        self.templateParameters = get_morpheus_terraform_inputs(self.custom_options)
        self.templateParameters["lab_name"] = self.instanceName.split("-")[-1]
        self.templateParameters["master_region"] = self.instance_map["customOptions"]["trainingZone"].split(",")[0]
        self.templateParameters["labs_region"] = self.instance_map["customOptions"]["trainingZone"].split(",")[1]
        self.instance_type = get_class_lab_instance_type(self.url,self.headers,f'ilt{self.instance_map["customOptions"]["class_name"][0:3]}')
        self.instance_layout_id = get_class_layout_id(self.url,self.headers,self.custom_options,self.instance_type)
        self.terraform_plan_id = get_morpheus_terraform_plan_id(self.url,self.headers)
        self.deploy_payload = get_instance_provisioning_payload(
            class_version=self.custom_options["class_version"],
            zone_id=self.instance_map["input"]["zoneId"],
            instance_name=self.instanceName,
            site_id=self.instance_map["input"]["site"]["id"],
            instance_type=self.instance_type["code"],
            instance_type_code=self.instance_type["code"],
            layout_id=self.instance_layout_id,
            plan_id=self.terraform_plan_id,
            template_parameters=self.templateParameters,
            )
        
    def deploy(self):
        try:
            logger.info(f'Attempting to deploy the instance: {self.instanceName}')
            resp = requests.post(f'{self.url}/api/instances',headers=self.headers,json=self.deploy_payload,verify=False)
        except Exception as e:
            logger.error(e)
        if "200" in str(resp):
            logger.info(f'Instance {self.instanceName} deployment successfully triggered...')
            self.instance_id = resp.json()["instance"]["id"]
            return(resp.text)
        elif "504" in str(resp):
            logger.info(f'Received a gateway timeout.')
            logger.info(f'Instance {self.instanceName} deployment possibly triggered...')
            self.instance_id = ""
            return(resp.text)
        else:
            logger.error(f'Something went wrong: {resp.json()}')
            return(resp.text)

    def status(self):
        counter = 0
        max_counter = 20
        logger.info(f'Checking the deployment status of the instance: {self.instanceName}')
        while counter < max_counter:
            if self.instance_id == "":
                logger.info(f'Instance ID not found. Attempting to get it via the instance name...')
                try:
                    resp = requests.get(f'{self.url}/api/instances/?name={self.instanceName}',headers=self.headers,verify=False)
                    counter += 1
                except Exception as e:
                    logger.error(f'Something went wrong: {e}')
                if "200" in str(resp):
                    logger.info(f'Instance {self.instanceName} instance found. Setting the ID to {resp.json()["instances"][0]["id"]}')
                    self.instance_id = resp.json()["instances"][0]["id"]
            else:
                logger.info(f'Instance ID found for instance {self.instanceName}. Checking status...')
                try:
                    resp = requests.get(f'{self.url}/api/instances/{self.instance_id}',headers=self.headers,verify=False)
                    counter += 1
                except Exception as e:
                    logger.error(f'Something went wrong: {e}')
                if "200" in str(resp):
                    logger.info(f'Status: {resp.json()["instance"]["status"]}')
                    if resp.json()["instance"]["status"] == "provisioning":
                        logger.info(f'Instance {self.instanceName} is still provisioning. Waiting 60 seconds and checking again...')
                        self.status = resp.json()["instance"]["status"]
                        time.sleep(60)
                    elif resp.json()["instance"]["status"] == "running":
                        logger.info(f'Instance {self.instanceName} is running. Continuing...')
                        self.status = resp.json()["instance"]["status"]
                        return
                    else:
                        logger.error(f'Unexpected status: {resp.json()["instance"]["status"]}')
                        self.status = resp.json()["instance"]["status"]
                        return
                else:
                    logger.error(f'Something went wrong: {resp.json()}')
                    return(resp.json())

    def lock(self):
        try:
            logger.info(f'Attempting to lock the instance: {self.instanceName}')
            resp = requests.put(f'{self.url}/api/instances/{self.instance_id}/lock',headers=self.headers,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        if "200" in str(resp):
            logger.info(f'Instance {self.instanceName} locked successfully.')
        else:
            logger.error(f'Something went wrong: {resp.json()}')
            return(resp.json())
        
    def get_lab_data(self):
        self.lab_data = {}
        try:
            logger.info(f'Attempting to get the lab info for the instance: {self.instanceName}')
            resp = requests.get(f'{self.url}/api/instances/{self.instance_id}/state',headers=self.headers,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        if "200" in str(resp):
            logger.info(f'Lab info for instance {self.instanceName} found.')
            if "administration" in self.class_name:
                logger.info(f'Class type: Administration')
                self.lab_data["lab_name"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "lab_name"][0]
                self.lab_data["region"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "labs_region"][0]
                self.lab_data["morpheus_external_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_external_ip"][0]
                self.lab_data["morpheus_internal_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_internal_ip"][0]
                self.lab_data["morpheus_url"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_url"][0]
                self.lab_data["rocky_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "rocky_ami"][0]
                self.lab_data["ubuntu_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "ubuntu_ami"][0]
                self.lab_data["centos_7_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "centos_7_ami"][0]
                self.lab_data["iam_key"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_key"][0]
                self.lab_data["iam_secret"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_secret"][0]
                self.lab_data["zion_vpc_id"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "zion_vpc_id"][0]
                self.lab_data["shared_vpc_id"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "shared_vpc_id"][0]
                return
            elif "automation" in self.class_name:
                logger.info(f'Class type: Automation')
                self.lab_data["lab_name"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "lab_name"][0]
                self.lab_data["region"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "labs_region"][0]
                self.lab_data["morpheus_external_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_external_ip"][0]
                self.lab_data["morpheus_internal_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_internal_ip"][0]
                self.lab_data["morpheus_url"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_url"][0]
                self.lab_data["rocky_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "rocky_ami"][0]
                self.lab_data["ubuntu_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "ubuntu_ami"][0]
                self.lab_data["centos_7_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "centos_7_ami"][0]
                self.lab_data["iam_key"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_key"][0]
                self.lab_data["iam_secret"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_secret"][0]
                self.lab_data["vpc_id"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "lab_vpc_id"][0]
                return
        else:
            logger.error(f'Something went wrong: {resp.json()}')
            return(resp.json())

class existingClass():
    def __init__(self,url,headers,instance_map) :
        self.url = url
        self.headers = headers
        self.instance_map = instance_map

    def show(self):
        logger.info(f'Instance map:')
        print(json.dumps(self.instance_map, indent = 4))

    def unlock(self):
        for lab in self.instance_map["attributes"]["labs"]:
            try:
                logger.info(f'Attempting to unlock the instance: {lab["lab_name"]}')
                resp = requests.put(f'{self.url}/api/instances/{lab["instance_id"]}/unlock',headers=self.headers,verify=False)
            except Exception as e:
                logger.error(f'Something went wrong: {e}')
            if "200" in str(resp):
                logger.info(f'Instance {lab["lab_name"]} unlocked successfully.')
            else:
                logger.error(f'Something went wrong: {resp.json()}')
                return(resp.json())


    def delete(self):
        for lab in self.instance_map["attributes"]["labs"]:
            try:
                logger.info(f'Attempting to delete the instance: {lab["lab_name"]}')
                resp = requests.delete(f'{self.url}/api/instances/{lab["instance_id"]}',headers=self.headers,verify=False)
            except Exception as e:
                logger.error(f'Something went wrong: {e}')
            if "200" in str(resp):
                logger.info(f'Instance {lab["lab_name"]} delete initiated successfully.')
                tries = 0
                maxTries = 10
                while tries < maxTries:
                    try:
                        tries += 1
                        ins = get_instance(self.url,self.headers,lab["instance_id"])
                    except Exception as e:
                        logger.error(f'Something went wrong: {e}')
                    if ins["status"] == "removing":
                        logger.info(f'Instance {lab["lab_name"]} still removing.')
                        logger.info(f'Waiting 60 seconds and checking again...')
                        time.sleep(60)
                        break
                    elif ins["status"] == "stopped":
                        logger.info(f'Instance {lab["lab_name"]} stopped but not yet deleted')
                        logger.info(f'Waiting 60 seconds and checking again...')
                        time.sleep(60)
                    elif "Instance not found" in str(ins):
                        logger.info(f'Instance not found. Assuming it is gone....')
                        self.instance_map["attributes"]["labs"].remove(lab)
                        return(self.instance_map)
                logger.info(f'Instance {lab["lab_name"]} not gone after 10 minutes. You may need to check on this')
            else:
                logger.error(f'Something went wrong: {resp.json()}')
                return(resp.json())

    def get_lab_data(self):
        self.lab_data = {}
        try:
            logger.info(f'Attempting to get the lab info for the instance: {self.instanceName}')
            resp = requests.get(f'{self.url}/api/instances/{self.instance_id}/state',headers=self.headers,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        if "200" in str(resp):
            logger.info(f'Lab info for instance {self.instanceName} found.')
            self.lab_data["lab_name"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "lab_name"][0]
            self.lab_data["region"] = [i["value"] for i in resp.json()["input"]["variables"] if i["name"] == "labs_region"][0]
            self.lab_data["morpheus_external_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_external_ip"][0]
            self.lab_data["morpheus_internal_ip"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_internal_ip"][0]
            self.lab_data["morpheus_url"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "morpheus_url"][0]
            self.lab_data["rocky_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "rocky_ami"][0]
            self.lab_data["ubuntu_ami"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "ubuntu_ami"][0]
            self.lab_data["iam_key"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_key"][0]
            self.lab_data["iam_secret"] = [i["value"]["value"] for i in resp.json()["output"]["outputs"] if i["name"] == "iam_secret"][0]
            return
        else:
            logger.error(f'Something went wrong: {resp.json()}')
            return(resp.json())

class existingLab():
    def __init__(self,lab_map) -> None:
        self.lab_map = lab_map
        self.url = self.lab_map["morpheus_url"]
        self.user = self.lab_map["ui_user"]
        self.password = self.lab_map["ui_password"]
        self.bearer = get_morpheus_bearer_token(self.url,self.user,self.password)
        self.headers = {'Content-Type': 'application/json',"Accept":"application/json", "Authorization": "BEARER " + self.bearer}

    def setup_admin_class(self):
        appliance_settings_body = {"applianceSettings":{"userBrowserSessionTimeout": "60","userBrowserSessionWarning": "15"}}
        provisioning_settings_body = {"provisioningSettings":{"showPricing": True,"reuseSequence": True,"cloudInitUsername": "morpheusci","cloudInitPassword": "Password123?"}}
        monitoring_settings_body = { "monitoringSettings": { "autoManageChecks": True } }
        logging_settings_body = { "logSettings": { "enabled": True } }
        try:
            logger.info(f'Attempting to update appliance settings')
            resp = requests.put(f'{self.url}/api/appliance-settings',headers=self.headers,json=appliance_settings_body,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        try:
            logger.info(f'Attempting to update provisioning settings')
            resp = requests.put(f'{self.url}/api/provisioning-settings',headers=self.headers,json=provisioning_settings_body,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        try:
            logger.info(f'Attempting to update monitoring settings')
            resp = requests.put(f'{self.url}/api/monitoring-settings',headers=self.headers,json=monitoring_settings_body,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')
        try:
            logger.info(f'Attempting to update logging settings')
            resp = requests.put(f'{self.url}/api/log-settings',headers=self.headers,json=logging_settings_body,verify=False)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')

        logger.info(f'Admin class setup complete')
        
    def setup_automation_class(self,**class_configs):
        logger.info(f'Begin attempt to setup Automation Class on node: {self.url}')
        # Ensure URL is https
        url = _validate_url(self.url)
        # Add the GIT integration
        integration = _add_git_integration_pat_auth(url,self.headers,class_configs["integration_name"],class_configs["git_url"],class_configs["git_branch"],class_configs["git_username"],class_configs["git_pat"])
        logger.info(f'Integration Data: {integration}')
        # Add the manual option list
        option_list_id = _add_manual_option_list(url,self.headers,class_configs["option_list_name"])
        logger.info(f'Option List ID: {option_list_id}')

        ## Create inputs
        input_ids = []
        input_ids.append(_add_input(self.url,self.headers,"AMI - CentOS", "amicentos", "text", "AMI - CentOS", "True"))
        input_ids.append(_add_input(self.url,self.headers,"AMI - Ubuntu", "amiubuntu", "text", "AMI - Ubuntu", "True"))
        input_ids.append(_add_input(self.url,self.headers,"AWS Key", "awsKey", "text", "AWS Key", "True"))
        input_ids.append(_add_input(self.url,self.headers,"AWS Secret", "awsSecret", "text", "AWS Secret", "True"))
        input_ids.append(_add_input(self.url,self.headers,"AWS VPC", "awsVpc", "text", "AWS VPC", "True"))
        input_ids.append(_add_input(self.url,self.headers,"Internal IP", "internalHost", "text", "Internal IP", "True"))
        input_ids.append(_add_select_input(self.url,self.headers,"AWS Regions", "awsRegion", option_list_id, "AWS Region", "True"))

        ## Create "Set Up My Lab" Python Task
        task_id = _add_python_task(self.url,self.headers,"Set Up My Lab", "setupMyLab", "12", "jythonTask", "value", "repository", integration["linked_code_repo_id"], "/setupmylab.py", "", "requests")

        ## Create "Set Up My Lab" Workflow
        workflow_id = _add_operational_workflow(self.url,self.headers,"Set Up My Lab", task_id, input_ids)

        ## Create catalog item
        catalog_id = _add_workflow_catalog_item(self.url,self.headers,"Set Up My Lab", workflow_id)

        ## Add logo
        logo = _add_catalog_logo(self.url,self.headers,catalog_id, "MorpheusImage.png")

class newLab():
    def __init__(self,url,headers,instance_map):
        self.url = url
        self.headers = headers
        self.instance_map = instance_map

    def setup(self):
        pass

class awsEnv():
    def __init__(self,instance_map) :
        self.instance_map = instance_map
        self.region = self.instance_map["attributes"]["labs"][0]["region"]
        access_key = self.instance_map["attributes"]["labs"][0]["iam_key"]
        secret_key = self.instance_map["attributes"]["labs"][0]["iam_secret"]
        self.ec2 = boto3.client('ec2',region_name=self.region,aws_access_key_id=access_key,aws_secret_access_key=secret_key)
        self.s3 = boto3.client('s3',region_name=self.region,aws_access_key_id=access_key,aws_secret_access_key=secret_key)
        self.rds = boto3.client('rds',region_name=self.region,aws_access_key_id=access_key,aws_secret_access_key=secret_key)
   
    def delete_s3_bucket_versions(self,bucket_name, object_versions):
        # Get all object versions
        logger.info(f'Attempting to delete all object versions in the bucket: {bucket_name}')
        versions = object_versions.get('Versions', [])
        if versions:
            delete_keys = {'Objects': [{'Key': version['Key'], 'VersionId': version['VersionId']} for version in versions]}
            try:
                resp = self.s3.delete_objects(Bucket=bucket_name, Delete=delete_keys)
                logger.info(resp)
            except Exception as e:
                logger.error(f'Something went wrong: {e}')
        
    def delete_s3_bucket_delete_markers(self,bucket_name, object_versions):
    # Get all object versions in the bucket's versioning-enabled delete marker
        logger.info(f'Attempting to delete all delete markers in the bucket: {bucket_name}')
        delete_markers = object_versions.get('DeleteMarkers', [])
        if delete_markers:
            # Delete all object versions in the bucket's versioning-enabled delete marker
            delete_marker_keys = {'Objects': [{'Key': marker['Key'], 'VersionId': marker['VersionId']} for marker in delete_markers]}
            try: 
                resp = self.s3.delete_objects(Bucket=bucket_name, Delete=delete_marker_keys)
                logger.info(resp)
            except Exception as e:
                logger.error(f'Something went wrong: {e}')

    def delete_s3_bucket(self,bucket_name):
        # Get all object versions
        logger.info(f'Attempting to delete the S3 bucket: {bucket_name}')
        object_versions = self.s3.list_object_versions(Bucket=bucket_name, MaxKeys=1000)
        while True:
            # Delete all object versions
            self.delete_s3_bucket_versions(bucket_name, object_versions)
            # Delete all delete markers
            self.delete_s3_bucket_delete_markers(bucket_name, object_versions)
            # Check if there are more object versions to delete
            if object_versions.get('IsTruncated'):
                object_versions = self.s3.list_object_versions(Bucket=bucket_name, MaxKeys=1000, KeyMarker=response['NextKeyMarker'], VersionIdMarker=response['NextVersionIdMarker'])
            else:
                break
        # Deleting S3 bucket
        self.s3.delete_bucket(bucket_name)

    def delete_db_instance(self,db_instance_identifier):
        logger.info(f'Attempting to delete the RDS instance: {db_instance_identifier}')
        try:
            self.rds.delete_db_instance(DBInstanceIdentifier=db_instance_identifier, SkipFinalSnapshot=True,DeleteAutomatedBackups=True)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')

    def delete_db_cluster(self,db_cluster_identifier):
        logger.info(f'Attempting to delete the RDS cluster: {db_cluster_identifier}')
        try:
            resp = self.rds.delete_db_cluster(DBClusterIdentifier=db_cluster_identifier, SkipFinalSnapshot=True)
            logger.info(resp)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')

    def terminate_ec2_instance(self,instance_id):
        logger.info(f'Attempting to terminate the EC2 instance: {instance_id}')
        try:
            resp = self.ec2.terminate_instances(InstanceIds=[instance_id])
            logger.info(resp)
        except Exception as e:
            logger.error(f'Something went wrong: {e}')

    def get_rds_clusters(self):
        rds_clusters = self.rds.describe_db_clusters()
        return(rds_clusters)
    
    def get_s3_buckets_in_region(self):
        bucket_list = []
        s3_buckets = self.s3.list_buckets()
        for bucket in s3_buckets["Buckets"]:
            if self.s3.get_bucket_location(Bucket=bucket["Name"])["LocationConstraint"] == self.region:
                bucket_list.append(bucket["Name"])
        return(bucket_list)
    
    def get_ec2_instances(self):
        ec2_instances = self.ec2.describe_instances()
        return(ec2_instances["Reservations"])


def get_morpheus_bearer_token(url,user_name,password):
    """
    The get_morpheus_api_token function is used to get a Morpheus API token.
    It takes three arguments: the URL of the Morpheus appliance, a username and password.
    The function returns an access_token which can be used in subsequent requests.
    
    :param url: Specify the morpheus api instance that is being used
    :param user_name: Specify the username of the user that is used to authenticate with the morpheus api
    :param password: Pass the password to get_morpheus_api_token function
    :return: A token for the user
    :doc-author: Trelent
    """
    if 'https://' not in url:
        url = f'https://{url}'
    endpoint = "/oauth/token?grant_type=password&scope=write&client_id=morph-api"
    headers = {'Content-Type': 'application/x-www-form-urlencoded'}
    body = {"username": user_name, "password": password}
    try:
        logger.info(f'Attempting to get API token')
        resp = requests.post(f'{url}{endpoint}',headers=headers,verify=False,data=body)
    except Exception as e:
        logger.info(f'Something went wrong')
    if "200" in str(resp):
        logger.info("Token acquired")
        token = resp.json()["access_token"]
        return(token)
    elif "400" in str(resp):
        logger.error('Bad credentials')
        return(resp)
    else:
        return(resp)


"""Setup for the chocobo package."""

from setuptools import setup, find_packages


# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    author="Sean Jabro",
    author_email="sean.jabro@outlook.com",
    name='molab',
    license="MIT",
    description='molab is a python package for configuring Morpheus training lab environments.',
    version='0.6.29',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/morpheus_training/molab',
    packages=find_packages() + find_packages(where='template_files'),
    package_dir={'template_files': 'template_files'},
    package_data={'molab': ['template_files/*.*']},
    python_requires=">=3.8",
    install_requires=["urllib3","requests","requests_futures","loguru","morpheus-cypher","importlib_resources","boto3"],
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
    ],
)